from typing import Any, Tuple, Union

import numpy as np
from scipy.sparse import csr_matrix

from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import MinMaxScaler
from sklearn.feature_extraction.text import (
    TfidfVectorizer,
    HashingVectorizer,
    CountVectorizer,
)
from sklearn.metrics.pairwise import (
    cosine_similarity,
    euclidean_distances,
    manhattan_distances,
    haversine_distances,
)

from .kb import QnAKnowledgeBase, FilePath, DEFAULT_KNOWLEDGE_BASE_FILE_PATH
from ._enums import EmbeddingModel, SimilarityMetric
from ._utils import value_error_message


class QnABot:
    _is_fitted: bool = False
    _model_kwargs: dict = {}

    _params = {"kb": None, "model": None, "ref_embeddings": None}

    def __init__(
        self,
        model_name: Union[str, EmbeddingModel] = "tfidf",
        similarity_metric: Union[str, SimilarityMetric] = "cosine",
        min_score: float = 0.25,
        cache: bool = False,
        **kwargs
    ) -> None:
        """Initializes an instance of the QnABot class.

        Args:
            model_name (Union[str, EmbeddingModel]): Name of the model used for text embedding. Defaults to 'tfidf'.
            similarity_metric (Union[str, SimilarityMetric]): Similarity metric used to find the most similar question.
                                                              Defaults to 'cosine'.
            min_score (float): Minimum similarity score below which an "I don't know" answer will be returned.
                               Defaults to 0.25.
            cache (bool): Whether to cache the entire knowledge base in memory. Defaults to False.
            **kwargs: Other keyword arguments supported to initialize models.

        """
        self.model_name: str = model_name
        self.similarity_metric: str = similarity_metric
        self.min_score: float = min_score
        self.cache: bool = cache

        self._model_kwargs = kwargs

    @staticmethod
    def _initialize_model(model_name: str, **kwargs):
        if model_name == "tfidf":
            return TfidfVectorizer(**kwargs)
        elif model_name == "murmurhash":
            return HashingVectorizer(**kwargs)
        elif model_name == "count":
            return CountVectorizer(**kwargs)
        else:
            raise ValueError(
                value_error_message(
                    "model_name", model_name, [e.value for e in EmbeddingModel]
                )
            )

    def fit(
        self, kb: Union[FilePath, QnAKnowledgeBase] = DEFAULT_KNOWLEDGE_BASE_FILE_PATH
    ):
        """Fits QnA Bot to a given knowledge base.

        Args:
            kb (Union[FilePath, QnAKnowledgeBase]): Path to the knowledge base JSON file or buffer, or a
                                                    QnAKnowledgeBase object. Defaults to the file path of the default
                                                    QnA Bot knowledge base.

        Returns:
            self: The instance itself.
        """
        self._is_fitted = False
        self._params["kb"] = (
            kb if isinstance(kb, QnAKnowledgeBase) else QnAKnowledgeBase(kb, self.cache)
        )
        self._params["model"] = self._initialize_model(
            model_name=self.model_name, **self._model_kwargs
        )
        self._params["model"].fit(self.knowledge_base_.ref_questions)
        self._params["ref_embeddings"] = self.model_.transform(
            self.knowledge_base_.ref_questions
        )

        self._is_fitted = True
        return self

    @property
    def _similarity_function(self):
        functions = {
            "cosine": cosine_similarity,
            "euclidean": euclidean_distances,
            "manhattan": manhattan_distances,
        }

        if self.similarity_metric not in functions:
            raise ValueError(
                value_error_message(
                    "similarity_metric",
                    self.similarity_metric,
                    [e.value for e in SimilarityMetric],
                )
            )
        else:
            return functions[self.similarity_metric]

    def find_similarity(self, input: str) -> Tuple[int, float]:
        """Returns the index and similarity score of the question in the knowledge base most similar to the input.

        Args:
            input (str): Input question.

        Returns:
            int: Index of the most similar question.
            float: Similarity score of the most similar question.
        """
        if not self._is_fitted:
            raise NotFittedError(
                "The model is not fitted. Use fit() method before calling answer()"
            )

        # Retrieve questions' indices from knowledge base
        q_idx = self.knowledge_base_.ref_questions_idx

        # Extract input statement embedding
        input_embeddings = self.model_.transform([input])

        # Calculate the similarities between the input embedding and the reference embeddings
        similarities = self._similarity_function(
            input_embeddings, self.ref_embeddings_
        ).flatten()

        if self.similarity_metric != "cosine":
            similarities = MinMaxScaler().fit_transform(similarities.reshape(-1, 1))
            similarities = 1.0 - similarities.flatten()

        # Find the score of the answer with the highest score
        highest_id = int(np.argmax(similarities))
        score = float(similarities[highest_id])

        # Find the ID of the answer with the highest score
        highest_qna_id = q_idx[highest_id]

        return highest_qna_id, score

    def answer(
        self, input: str, return_score: bool = False
    ) -> Union[str, Tuple[str, float]]:
        """Returns the index and similarity score of a question in the knowledge base that is most similar to the input.

        Args:
            input (str): Input question.
            return_score (bool): Whether to return the similarity score. Defaults to False.

        Returns:
            Union[str, Tuple[str, float]]: One of the answers of the most similar question in the knowledge base
                                           if return_score=False, otherwise a tuple containing:
                str: One of the answers of the most similar question in the knowledge base.
                float: Similarity score.
        """
        highest_id, score = self.find_similarity(input)

        # If score was lower than the minimum accepted value
        if score < self.min_score:
            # Return a random "I don't know" answer
            answer_ = np.random.choice(self.knowledge_base_.idk_answers)

        else:
            # Pick a random answer among the answers of the most similar question
            answer_ = np.random.choice(self.knowledge_base_.qna[highest_id]["a"])

        if return_score:
            return answer_, score
        else:
            return answer_

    @property
    def knowledge_base_(self) -> QnAKnowledgeBase:
        """Returns the knowledge base on which the QnA Bot is fitted.

        """
        return self._params["kb"]

    @property
    def model_(self) -> Any:
        """Returns the embedding model fitted on the knowledge base.

        """
        return self._params["model"]

    @property
    def ref_embeddings_(self) -> csr_matrix:
        """Returns the embedding matrix extracted from reference questions in the knowledge base.

        """
        return self._params["ref_embeddings"]

    def __repr__(self):
        return (
            "<QnABot(model_name='%s', similarity_metric='%s', min_score=%.2f, cache=%s)>"
            % (
                str(self.model_name),
                str(self.similarity_metric),
                self.min_score,
                self.cache,
            )
        )



"""
QnA Builder is a similarity-based conversational dialog engine for Python that helps you quickly and easily create
smart chatbots using a knowledge base of pre-defined questions and answers.
"""

__version__ = "0.1.2"
__author__ = "Mehdi Samsami"

__all__ = (
    "QnABot",
    "EmbeddingModel",
    "SimilarityMetric",
    "QnAKnowledgeBase",
    "DEFAULT_KNOWLEDGE_BASE_FILE_PATH",
)


from .qna_bot import QnABot
from ._enums import EmbeddingModel, SimilarityMetric
from .kb import QnAKnowledgeBase, DEFAULT_KNOWLEDGE_BASE_FILE_PATH



from typing import Any


def parse_list_options(options: list) -> str:
    """Returns the parsed version of a list of options in string.

    Args:
        options (list): A list of options, parameters, or items to be parsed.

    Returns:
        str: Comma-separated sequence of options with an 'or' after the last comma.
    """
    check_type_error("options", options, list())

    options_str = ", ".join([str(option) for option in options])
    last_comma_idx = options_str.rfind(",")

    if len(options) == 2:
        return options_str[:last_comma_idx] + " or" + options_str[last_comma_idx + 1 :]
    else:
        return (
            options_str[: last_comma_idx + 2] + "or" + options_str[last_comma_idx + 1 :]
        )


def value_error_message(name: str, value: Any, options: list) -> str:
    if len(options) == 1:
        return f"{value} is not a valid {name}. Only {options[0]} is supported"
    else:
        return f"{value} is not a valid {name}. Must be either {parse_list_options(options)}"


def type_error_message(name: str, value: Any, acceptable: Any) -> str:
    return f"{name} must be of type {type(acceptable).__name__}, {type(value).__name__} was given instead"


def check_value_error(name: str, value: Any, options: list):
    if value not in options:
        raise ValueError(value_error_message(name, value, options))


def check_type_error(name: str, value: Any, acceptable: Any):
    if not isinstance(value, type(acceptable)):
        raise TypeError(type_error_message(name, value, acceptable))



from unittest import TestCase

from qnabuilder import QnABot, EmbeddingModel, SimilarityMetric


class TestQnABot(TestCase):
    def test_wrong_model_name(self):
        bot = QnABot(model_name="tensorflow")
        with self.assertRaises(ValueError) as ctx:
            bot.fit()
        self.assertIn("tensorflow is not a valid model_name", str(ctx.exception))

    def test_wrong_similarity_metric(self):
        bot = QnABot(similarity_metric="mahalanobis")
        with self.assertRaises(ValueError) as ctx:
            bot.fit()
            bot.answer("hi")
        self.assertIn("mahalanobis is not a valid similarity_metric", str(ctx.exception))

    def test_generic_answer_tfidf(self):
        for metric in SimilarityMetric:
            bot = QnABot(model_name=EmbeddingModel.TFIDF, similarity_metric=metric)
            bot.fit()
            self.assertEqual(bot.answer("Who are you?"), "I am QnA Builder!")

    def test_generic_answer_murmurhash(self):
        for metric in SimilarityMetric:
            bot = QnABot(model_name=EmbeddingModel.MURMURHASH, similarity_metric=metric)
            bot.fit()
            self.assertEqual(bot.answer("Do you have a name?"), "I am QnA Builder!")
    
    def test_generic_answer_count(self):
        for metric in SimilarityMetric:
            bot = QnABot(model_name=EmbeddingModel.COUNT, similarity_metric=metric)
            bot.fit()
            self.assertEqual(bot.answer("So what's your name?"), "I am QnA Builder!")



import codecs
from os import path
from setuptools import setup


setup(
    name="qna-builder",
    version="0.1.3",
    description="Similarity-based conversational dialog engine for Python.",
    keywords=[
        "python",
        "qna",
        "qnabuilder",
        "chat",
        "chatbot",
        "conversation",
        "dialog",
    ],
    author="Mehdi Samsami",
    author_email="mehdisamsami@live.com",
    url="https://github.com/msamsami/qna-builder",
    long_description=codecs.open(
        path.join(path.abspath(path.dirname(__file__)), "README.md"), encoding="utf-8"
    ).read(),
    long_description_content_type="text/markdown",
    packages=["qnabuilder"],
    classifiers=[
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Linguistic",
        "Topic :: Communications :: Chat",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.7",
    install_requires=["scikit-learn==1.0.2"],
    extras_require={
        "dev": ["streamlit==1.21.0", "pytest==7.3.1"],
        "editor": ["streamlit==1.21.0"],
    },
)



from enum import Enum


__all__ = ["EmbeddingModel", "SimilarityMetric"]


class EmbeddingModel(str, Enum):
    """
    Names of text embedding models.
    """

    TFIDF = "tfidf"
    MURMURHASH = "murmurhash"
    COUNT = "count"


class SimilarityMetric(str, Enum):
    """
    Names of similarity calculation metrics.
    """

    COSINE = "cosine"
    EUCLIDEAN = "euclidean"
    MANHATTAN = "manhattan"
